import os
import unittest

from fictus import FictusFileSystem
from fictus.fictusexception import FictusException


class MyTestCase(unittest.TestCase):
    root = "c:"

    def setUp(self) -> None:
        self.fs = FictusFileSystem(self.root)

    def test_cd_back_one(self):
        self.fs.mkdir(r"a\b")  # create \\a\b
        self.fs.cd(r"\a\b")  # move to \\a\b

        # Should show c:\a\b
        self.assertEqual(os.sep.join([self.root, "a", "b"]), self.fs.cwd())

        self.fs.cd("..")  # Go higher
        self.assertEqual(os.sep.join([self.root, "a"]), self.fs.cwd())

        self.fs.cd("..")  # again to _root
        self.assertEqual(os.sep.join([self.root]), self.fs.cwd())

        with self.assertRaises(FictusException):
            # should fail as moving higher than _root doesn't exist
            self.fs.cd("..")

    def test_cd_traversal(self):
        self.fs.mkdir("a/b/c")
        self.fs.mkdir("a/c/b")
        self.fs.mkdir("a/a/a")

        self.fs.cd("a/b/c/../..")
        self.assertEqual(os.sep.join([self.root, "a"]), self.fs.cwd())

        self.fs.cd("/a/a/../b/../c")
        self.assertEqual(os.sep.join([self.root, "a", "c"]), self.fs.cwd())

    def test_cd_root(self):
        self.fs.mkdir("a/b")
        self.fs.cd("a/../a/b")
        self.fs.cd("/")
        self.assertEqual(self.root, self.fs.cwd())

        self.fs.cd("a/b")
        self.fs.cd("\\")
        self.assertEqual(self.root, self.fs.cwd())

        # go to _root from _root
        self.fs.cd("/a")
        self.fs.cd("/a")
        self.assertEqual(os.sep.join([self.root, "a"]), self.fs.cwd())

    def test_cd_from_cwd(self):
        self.fs.mkdir("a/b/c")
        self.fs.mkdir("a/c/b")
        self.fs.mkdir("a/a/a")
        self.fs.cd("/a/b")

        self.assertEqual(os.sep.join([self.root, "a", "b"]), self.fs.cwd())

        self.fs.cd("/a")
        self.assertEqual(os.sep.join([self.root, "a"]), self.fs.cwd())

    def test_cd_fail(self):
        self.fs.cd("/")  # drop to _root

        with self.assertRaises(FictusException):
            self.fs.cd("z/y/x")  # doesn't go anywhere; invalid path
            self.fs.cd("/z/y/x")  # still broken but starts with a jump to _root

    def test_cd_fail_but_remain_at_cwd(self):
        self.fs.cd("/")  # drop to _root
        self.fs.mkdir("a/b/c")
        self.fs.cd("a/b/c")  # move down the chain

        with self.assertRaises(FictusException):
            self.fs.cd("../../../..")  # go back once too many times

        # still at the last place expected
        self.assertEqual(self.fs.cwd(), self.root + "/a/b/c")

    def test_mkdir_with_non_string(self):
        with self.assertRaises(FictusException):
            # empty strings not allowed
            self.fs.mkdir("")

    def test_mkdir_with_odd_characters(self):
        # allowed
        items = [" ", "0", "~", "🎂"]
        for item in items:
            self.fs.mkdir(f"/{item}")
            self.fs.cd(f"/{item}")
            self.assertEqual(os.sep.join([self.root, item]), self.fs.cwd())

    def test_mkdir_with_a_dot(self):
        # a period is 'normalized' away ie: fizz/./buzz == fizz/buzz
        self.fs.mkdir("/fizz/./buzz")
        self.fs.cd("/fizz/buzz")  # no '.' directory created
        self.fs.cd("/fizz/./././././././././.")  # still fine.
        self.assertEqual(os.sep.join([self.root, "fizz"]), self.fs.cwd())


if __name__ == "__main__":
    unittest.main()



import unittest

from fictus.folder import Folder


class MyTestCase(unittest.TestCase):
    def test_parent(self):
        a = Folder("a")
        b = Folder("b")
        a.folder(b)
        self.assertEqual(
            a,
            b.parent,
        )
        self.assertEqual(None, a.parent)


if __name__ == "__main__":
    unittest.main()



import os.path
from typing import Set, Optional

from fictus.data import Data

from .fictusexception import FictusException
from .file import File
from .folder import Folder


DEFAULT_ROOT_NAME = "\\"


class FictusFileSystem:
    """
    A FictusFileSystem (FFS) simulates the creation and traversal of a file system.
    The FFS allows for the creation and removal of files and folders,
    """

    def __init__(self, name=DEFAULT_ROOT_NAME) -> None:
        self._root: Folder = Folder(name)
        self._current: Folder = self._root

    def current(self) -> Folder:
        return self._current

    @staticmethod
    def _normalize(path: str) -> str:
        return os.path.normpath(path.replace(DEFAULT_ROOT_NAME, "/"))

    def mkdir(self, path: str) -> None:
        """Takes a string of a normalized relative to cwd and adds the directories
        one at a time."""
        if not path:
            raise FictusException("A Folder must contain a non-empty string.")

        # hold onto the current directory
        current = self._current

        normalized_path = self._normalize(path)
        if normalized_path.startswith(os.sep):
            self._to_root()

        folders = {d.name: d for d in self._current.folders()}

        for part in normalized_path.split(os.sep):
            if not part:
                continue

            if part not in folders:
                folders[part] = Folder(part)
                self._current.folder(folders[part])

            self.cd(folders[part].name)
            folders = {d.name: d for d in self._current.folders()}

        # return to starting directory
        self._current = current

    def mkfile(self, *files: str) -> None:
        """Takes one or more filenames and adds them to the cwd."""
        visited: Set[str] = {f.name for f in self._current.files()}
        for file in files:
            if not file:
                raise FictusException("A File must contain a non-empty string.")

            if file not in visited:
                visited.add(file)
                self._current.file(File(file))

    def rename(self, old: str, new: str) -> None:
        """Renames a File or Folder based on its name."""
        for content in self._current.contents():
            if content.name == old:
                content.name = new
                break

    def cwd(self) -> str:
        """Prints the current working directory."""
        r = []

        node: Optional[Data] = self._current
        while node is not None:
            r.append(node.name)
            node = node.parent

        return f"{os.sep}".join(reversed(r))

    def _to_root(self) -> None:
        self._current = self._root

    def cd(self, path: str) -> None:
        """Takes a string of a normalized relative to cwd and changes the current"""
        # Return to the current dir if something goes wrong
        current = self._current

        normalized_path = self._normalize(path)
        if normalized_path.startswith(os.sep):
            self._to_root()

        for index, part in enumerate(normalized_path.split(os.sep)):
            if not part:
                continue

            if index == 0 and part == self._root.name:
                self._to_root()
                continue

            if part == "..":
                # looking at the parent here, so ensure its valid.
                parent = self._current.parent

                if parent is None or isinstance(parent, Folder) is False:
                    self._current = current
                    raise FictusException(
                        f"Could not path to {normalized_path} from {self.cwd()}."
                    )

                self._current = parent

            else:
                hm = {f.name: f for f in self._current.folders()}
                if part not in hm:
                    self._current = current
                    raise FictusException(
                        f"Could not path to {normalized_path} from {self.cwd()}."
                    )
                self._current = hm[part]

        return None



from __future__ import annotations

import os
import re
import sys
from typing import List, Set, Optional

from .fictusfilesystem import FictusFileSystem
from .constants import PIPE, SPACER_PREFIX, ELBOW, TEE, SPACER
from .data import Data
from .file import File
from .folder import Folder
from .renderer import Renderer, defaultRenderer

pattern = re.compile(r"[^\\]")


class FictusDisplay:
    def __init__(self, ffs: FictusFileSystem):
        self._ffs = ffs
        self._renderer = defaultRenderer
        self._ignore: Set[int] = set()

    def set_renderer(self, renderer: Renderer) -> None:
        self._renderer = renderer

    def _display_node(self, node: Data) -> str:
        """
        Bookkeeping of nested node depth, node siblings, and order in the queue are
        used to present the FicusSystem in an aesthetic way.
        """

        parts = [PIPE + SPACER_PREFIX for _ in range(node.level)]
        for index in self._ignore:
            if len(parts) > index - 1:
                parts[index - 1] = SPACER + SPACER_PREFIX

        if parts:
            parts[-1] = ELBOW if node.last is True else TEE

        is_file = isinstance(node, File)
        file_open = self._renderer.file_open if is_file else self._renderer.folder_open
        file_close = (
            self._renderer.file_close if is_file else self._renderer.folder_close
        )

        # checking for Folder type
        end = "\\" if not is_file else ""

        return f'{"".join(parts)}{file_open}{node.name}{file_close}{end}'

    @staticmethod
    def _pprint_header(header: str) -> int:
        """Writes the CWD to stdout with forward slashes and its length."""

        parts = header.split(os.sep)
        if len(parts) <= 1:
            # when _root is passed in
            return 0

        header = f"\\".join(parts[:-1])
        sys.stdout.write(f"{header}\\\n")
        return header[:-1].rfind("\\") + 1  # one past found

    def pprint(self, renderer: Optional[Renderer] = None) -> None:
        """Displays the file system structure to stdout."""

        node = self._ffs.current()
        node.last = True
        self._ignore = {i for i in range(node.level)}
        header_length = self._pprint_header(self._ffs.cwd())

        prefix: int = -1  # not set
        buffer: List[str] = [self._renderer.doc_open]

        q: List[Data] = [node]
        while q:
            node = q.pop()
            if node.last is False:
                if node.level in self._ignore:
                    self._ignore.remove(node.level)
            line = self._display_node(node)

            # This needs to happen only once and applied
            # thereafter to each subsequent line.
            prefix = len(line) - len(line.lstrip()) if prefix == -1 else prefix

            buffer.append(f"{header_length * SPACER}{line[prefix:]}\n")

            if node.last is True:
                # track nodes without children.
                self._ignore.add(node.level)

            if isinstance(node, Folder):
                q += node.contents()

            # clear flag for next run
            node.last = False
        buffer.append(self._renderer.doc_close)

        sys.stdout.writelines(buffer)



from __future__ import annotations
from typing import List, Union, Sequence, Optional

from .data import Data
from .file import File


class Folder(Data):
    """
    A Folder is a data item in the FictusSystem that may contain other Folders
    and/or Files.
    """

    def __init__(self, name: str, parent: Optional[Folder] = None):
        super().__init__(name, parent)

        self._folders: List[Folder] = []
        self._files: List[File] = []

    def file(self, file: File) -> None:
        """Adds a file to the current Folder."""
        file.level = self._level + 1
        self._files.append(file)

    def files(self) -> List[File]:
        """Returns an alphabetized list of files found in the current Folder."""
        return sorted(self._files[::])

    def folder(self, folder: Folder) -> None:
        """Adds a direct sub-folder to the current Folder."""
        folder.parent = self
        folder._level = self._level + 1
        self._folders.append(folder)

    def folders(self) -> List[Folder]:
        """Returns an alphabetized list of folders found in the current Folder."""
        return sorted(self._folders[::])

    def contents(self) -> Sequence[Union[File, Folder]]:
        """Returns an alphabetized list of folders and files found in the current Folder."""
        items: List[Union[File, Folder]] = []
        items += sorted(self._files[::])
        items += sorted(self._folders[::])
        if items:
            items[0].last = True
        return items



from dataclasses import field, dataclass

"""A Renderer is a representation of how a fs should be printed."""


@dataclass
class Renderer:
    """A Renderer provides special instructions for how a fs is displayed."""

    doc_open: str = field(default="")
    doc_close: str = field(default="")
    file_open: str = field(default="")
    file_close: str = field(default="")
    folder_open: str = field(default="")
    folder_close: str = field(default="")


defaultRenderer = Renderer()

markdownRenderer = Renderer(
    '<pre style="line-height:17px">',
    "</pre>",  # Doc open/close
    '<span style="color:gray">',
    "</span>",  # File open/close
)

emojiRenderer = Renderer(
    "",
    "",  # Doc open/close
    "📄",
    "",  # File open/close
    "📁",
    "",  # Folder open/close
)



from __future__ import annotations

from typing import Optional


class Data:
    def __init__(self, name: str, parent: Optional[Data] = None):
        self.name = name
        self.last = False

        self._level = 0
        self._parent = parent

    @property
    def level(self) -> int:
        return self._level

    @level.setter
    def level(self, val: int) -> None:
        self._level = val

    @property
    def parent(self) -> Optional[Data]:
        return self._parent

    @parent.setter
    def parent(self, other: Optional[Data]) -> None:
        self._parent = other

    def __lt__(self, other: Data) -> bool:
        return self.name > other.name
